import asyncio
from typing import Dict

import mongomock as mongomock
from mongoengine import connect
from fastapi import FastAPI

from fast_agave.middlewares import FastAgaveErrorHandler
from .resources import app as resources
from .middlewares import AuthedMiddleware
from .tasks.task_example import dummy_task

connect(
    host='mongodb://localhost:27017/db',
    mongo_client_class=mongomock.MongoClient,
)
app = FastAPI(title='example')
app.include_router(resources)

app.add_middleware(AuthedMiddleware)
app.add_middleware(FastAgaveErrorHandler)


@app.get('/')
async def iam_healty() -> Dict:
    return dict(greeting="I'm healthy!!!")


@app.on_event('startup')
async def on_startup() -> None:  # pragma: no cover
    # Inicializa el task que recibe mensajes
    # provenientes de SQS
    asyncio.create_task(dummy_task())



from fast_agave.tasks.sqs_tasks import task


# Esta URL es solo un mock de la queue.
# Debes reemplazarla con la URL de tu queue
QUEUE_URL = 'http://127.0.0.1:4000/123456789012/core.fifo'


@task(queue_url=QUEUE_URL, region_name='us-east-1')
async def dummy_task(message) -> None:
    print(message)



from cuenca_validations.errors import CuencaError
from fastapi import Request, Response
from fastapi.responses import JSONResponse
from fastapi.routing import APIRoute
from starlette.middleware.base import (
    BaseHTTPMiddleware,
    RequestResponseEndpoint,
)
from starlette.routing import Match

from ..exc import FastAgaveError, MethodNotAllowedError, NotFoundError


class FastAgaveErrorHandler(BaseHTTPMiddleware):
    async def dispatch(
        self, request: Request, call_next: RequestResponseEndpoint
    ) -> Response:
        try:
            request.scope['route_handler'] = get_current_route_handler(request)
            return await call_next(request)
        except CuencaError as exc:
            return JSONResponse(
                status_code=exc.status_code,
                content=dict(
                    code=exc.code,
                    error=str(exc),
                ),
            )
        except FastAgaveError as exc:
            return JSONResponse(
                status_code=exc.status_code, content=dict(error=exc.error)
            )


def get_current_route_handler(request: Request) -> APIRoute:
    """
    Helper method for getting the route handler of the current request.

    If there is not route handler it raises appropriate status code error
    consistent with the `Route.__call__` behavior
    https://github.com/encode/starlette/blob/5d768322d6d7adc31df54b1ad306f417e3da2c81/starlette/routing.py#L656-L666
    Args:
        request: fastapi request object

    Returns:
        APIRoute instance for the current request
    """
    partial = None
    for route in request.app.routes:
        match, _ = route.matches(request.scope)
        if match == Match.FULL:
            return route
        if match == Match.PARTIAL and partial is None:
            partial = route

    if partial is not None:
        raise MethodNotAllowedError('Method Not Allowed')
    else:
        raise NotFoundError('Not Found')



from dataclasses import dataclass


@dataclass
class FastAgaveError(Exception):
    error: str
    status_code: int


@dataclass
class BadRequestError(FastAgaveError):
    status_code: int = 400


@dataclass
class UnauthorizedError(FastAgaveError):
    status_code: int = 401


@dataclass
class ForbiddenError(FastAgaveError):
    status_code: int = 403


@dataclass
class NotFoundError(FastAgaveError):
    status_code: int = 404


@dataclass
class MethodNotAllowedError(FastAgaveError):
    status_code: int = 405


@dataclass
class ConflictError(FastAgaveError):
    status_code: int = 409


@dataclass
class UnprocessableEntity(FastAgaveError):
    status_code: int = 422


@dataclass
class TooManyRequests(FastAgaveError):
    status_code: int = 429


@dataclass
class FastAgaveViewError(FastAgaveError):
    status_code: int = 500


class RetryTask(Exception):
    ...



from functools import wraps

from fast_agave.blueprints.decorators import copy_attributes


def i_am_test(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    wrapper.i_am_test = True
    return wrapper


class TestResource:
    @i_am_test
    def retrieve(self) -> str:
        return 'hello'


def test_copy_properties_from() -> None:
    def retrieve():
        ...

    assert not hasattr(retrieve, 'i_am_test')
    retrieve = copy_attributes(TestResource)(retrieve)
    assert hasattr(retrieve, 'i_am_test')



from fastapi import Request
from fastapi.responses import JSONResponse as Response

from ..models.transactions import Transaction as TransactionModel
from .base import app


@app.resource('/transactions')
class Transaction:
    model = TransactionModel

    @staticmethod
    async def create(request: Request) -> Response:
        return Response(content={}, status_code=201)



__all__ = ['Account', 'Biller', 'Card', 'Transaction', 'File', 'User']

from .accounts import Account
from .billers import Biller
from .cards import Card
from .files import File
from .transactions import Transaction
from .users import User



import json
from base64 import b64encode
from dataclasses import dataclass
from typing import Dict, Iterable, Optional
from uuid import uuid4

from aiobotocore.session import get_session


def _build_celery_message(
    task_name: str, args_: Iterable, kwargs_: Dict
) -> str:
    task_id = str(uuid4())
    # la definición de esta plantila se encuentra en:
    # docs.celeryproject.org/en/stable/internals/protocol.html#definition
    message = dict(
        properties=dict(
            correlation_id=task_id,
            content_type='application/json',
            content_encoding='utf-8',
            body_encoding='base64',
            delivery_info=dict(exchange='', routing_key='celery'),
        ),
        headers=dict(
            lang='py',
            task=task_name,
            id=task_id,
            root_id=task_id,
            parent_id=None,
            group=None,
        ),
        body=_b64_encode(
            json.dumps(
                (
                    args_,
                    kwargs_,
                    dict(
                        callbacks=None, errbacks=None, chain=None, chord=None
                    ),
                )
            )
        ),
    )
    message['content-encoding'] = 'utf-8'
    message['content-type'] = 'application/json'

    encoded = _b64_encode(json.dumps(message))
    return encoded


def _b64_encode(value: str) -> str:
    encoded = b64encode(bytes(value, 'utf-8'))
    return encoded.decode('utf-8')


@dataclass
class SqsCeleryClient:
    queue_url: str
    region_name: str

    async def send_task(
        self,
        name: str,
        args: Optional[Iterable] = None,
        kwargs: Optional[Dict] = None,
    ) -> None:
        celery_message = _build_celery_message(name, args or (), kwargs or {})
        session = get_session()
        async with session.create_client('sqs', self.region_name) as sqs:
            await sqs.send_message(
                QueueUrl=self.queue_url,
                MessageBody=celery_message,
                MessageGroupId=str(uuid4()),
            )



from mongoengine import DateTimeField, StringField
from mongoengine_plus.aio import AsyncDocument
from mongoengine_plus.models import BaseModel

from mongoengine_plus.models.helpers import uuid_field


class Account(BaseModel, AsyncDocument):
    id = StringField(primary_key=True, default=uuid_field('AC'))
    name = StringField(required=True)
    user_id = StringField(required=True)
    platform_id = StringField(required=True)
    created_at = DateTimeField()
    deactivated_at = DateTimeField()



import mimetypes
from typing import Any, Dict, List, Optional
from urllib.parse import urlencode

from cuenca_validations.types import QueryParams
from fastapi import APIRouter, BackgroundTasks, Depends, Request, status
from fastapi.responses import JSONResponse as Response
from fastapi.responses import StreamingResponse
from mongoengine import DoesNotExist, Q
from pydantic import ValidationError
from pydantic.main import BaseConfig, BaseModel
from starlette_context import context

from ..exc import NotFoundError, UnprocessableEntity
from .decorators import copy_attributes

SAMPLE_404 = {
    "summary": "Not found item",
    "value": {"error": "Not valid id"},
}


class RestApiBlueprint(APIRouter):
    @property
    def current_user_id(self) -> str:
        return context['user_id']

    @property
    def current_platform_id(self) -> str:
        return context['platform_id']

    def user_id_filter_required(self) -> bool:
        return context['user_id_filter_required']

    def platform_id_filter_required(self) -> bool:
        return context['platform_id_filter_required']

    def custom_filter_required(self, query_params: Any, model: Any) -> None:
        """
        Overwrite this method in order to add new context
        based on custom filter.
        set de value of your filter ex query_params.wallet = self.wallet
        """
        pass

    async def retrieve_object(
        self, resource_class: Any, resource_id: str
    ) -> Any:
        resource_id = (
            self.current_user_id if resource_id == 'me' else resource_id
        )
        query = Q(id=resource_id)
        if self.platform_id_filter_required() and hasattr(
            resource_class.model, 'platform_id'
        ):
            query = query & Q(platform_id=self.current_platform_id)

        if self.user_id_filter_required() and hasattr(
            resource_class.model, 'user_id'
        ):
            query = query & Q(user_id=self.current_user_id)

        try:
            data = await resource_class.model.objects.async_get(query)
        except DoesNotExist:
            raise NotFoundError('Not valid id')
        return data

    def resource(self, path: str):
        """Decorator to transform a class in FastApi REST endpoints

        @app.resource('/my_resource')
        class Items(Resource):
            model = MyMongoModel
            response_model = MyPydanticModel (Resource Interface)
            query_validator = MyPydanticModel

            def create(): ...
            def delete(id): ...
            def retrieve(id): ...
            def get_query_filter(): ...

        This implementation create the following endpoints

        POST /my_resource
        PATCH /my_resource
        DELETE /my_resource/id
        GET /my_resource/id
        GET /my_resource
        """

        def wrapper_resource_class(cls):
            """Wrapper for resource class
            :param cls: Resoucre class
            :return:
            """
            response_model = Any
            response_sample = {}
            include_in_schema = getattr(cls, 'include_in_schema', True)
            if hasattr(cls, 'response_model'):
                response_model = cls.response_model
                response_sample = response_model.schema().get('example')

            """ POST /resource
            Create a FastApi endpoint using the method "create"

            OR using the method "upload" to enable POST using a
            streaming multipart parser to receive files as form data. It
            validates form data using `Resource.upload_validator`.
            """
            if hasattr(cls, 'create'):
                route = self.post(
                    path,
                    summary=f'{cls.__name__} - Create',
                    response_model=response_model,
                    status_code=status.HTTP_201_CREATED,
                    include_in_schema=include_in_schema,
                )
                route(cls.create)
            elif hasattr(cls, 'upload'):

                @self.post(
                    path,
                    summary=f'{cls.__name__} - Upload',
                    response_model=response_model,
                    include_in_schema=include_in_schema,
                    openapi_extra={
                        "requestBody": {
                            "content": {
                                "form-data": {
                                    "schema": cls.upload_validator.schema()
                                }
                            }
                        }
                    },
                )
                @copy_attributes(cls)
                async def upload(
                    request: Request, background_tasks: BackgroundTasks
                ):
                    form = await request.form()
                    try:
                        upload_params = cls.upload_validator(**form)
                    except ValidationError as exc:
                        return Response(content=exc.json(), status_code=400)

                    return await cls.upload(upload_params, background_tasks)

            """ DELETE /resource/{id}
            Use "delete" method (if exists) to create the FastApi endpoint
            """
            error_404 = json_openapi(404, 'Item not found', [SAMPLE_404])
            if hasattr(cls, 'delete'):

                @self.delete(
                    path + '/{id}',
                    summary=f'{cls.__name__} - Delete',
                    response_model=response_model,
                    responses=error_404,
                    description=f'Use id param to delete the {cls.__name__} object',
                    include_in_schema=include_in_schema,
                )
                @copy_attributes(cls)
                async def delete(id: str, request: Request):
                    obj = await self.retrieve_object(cls, id)
                    return await cls.delete(obj, request)

            """ PATCH /resource/{id}
            Enable PATCH method if Resource.update method exist. It validates
            body data using `Resource.update_validator` but update logic is
            completely your responsibility.
            """
            if hasattr(cls, 'update'):

                @self.patch(
                    path + '/{id}',
                    summary=f'{cls.__name__} - Update',
                    response_model=response_model,
                    responses=error_404,
                    description=f'Use id param to update the {cls.__name__} object',
                    include_in_schema=include_in_schema,
                )
                @copy_attributes(cls)
                async def update(
                    id: str,
                    update_params: cls.update_validator,  # type: ignore
                    request: Request,
                ):
                    obj = await self.retrieve_object(cls, id)
                    try:
                        return await cls.update(obj, update_params, request)
                    except TypeError:
                        return await cls.update(obj, update_params)

            """ GET /resource/{id}
            By default GET method only fetch object from DB.
            If you need extra logic override "retrieve" or "download" methods
            """

            @self.get(
                path + '/{id}',
                summary=f'{cls.__name__} - Retrieve',
                response_model=response_model,
                responses=error_404,
                description=f'Use id param to retrieve the {cls.__name__} object',
                include_in_schema=include_in_schema,
            )
            @copy_attributes(cls)
            async def retrieve(id: str, request: Request):
                """GET /resource/{id}
                :param id: Object Id
                :return: Model object

                If exists "retrieve" method return the result of that, else
                use "id" param to retrieve the object of type "model" defined
                in the decorated class.

                The most of times this implementation is enough and is not
                necessary define a custom "retrieve" method
                """
                obj = await self.retrieve_object(cls, id)

                # This case is when the return is not an application/$
                # but can be some type of file such as image, xml, zip or pdf
                if hasattr(cls, 'download'):
                    file = await cls.download(obj)
                    mimetype = request.headers['accept']
                    extension = mimetypes.guess_extension(mimetype)
                    filename = f'{cls.model._class_name}.{extension}'
                    result = StreamingResponse(
                        file,
                        media_type=mimetype,
                        headers={
                            'Content-Disposition': f'attachment; filename={filename}'
                        },
                    )
                elif hasattr(cls, 'retrieve'):
                    result = await cls.retrieve(obj)
                else:
                    result = obj.to_dict()

                return result

            """ GET /resource?param=value
            Use GET method to fetch and count filtered objects using query params.
            To Enable queries you have to define next fields in decorated class

            query_validator: Pydantic model to validate the params.
            get_query_filter: Method to provide the way that the params are used to filter data.
            """

            if not hasattr(cls, 'query_validator') or not hasattr(
                cls, 'get_query_filter'
            ):
                return cls

            query_description = (
                f'Make queries in resource {cls.__name__} and filter the result using query parameters.  \n'
                f'The items are paginated, to iterate over them use the `next_page_uri` included in response.  \n'
                f'If you need only a counter not the data send value `true` in `count` param.'
            )

            # Build dynamically types for query response
            class QueryResponse(BaseModel):
                items: Optional[List[response_model]] = []
                next_page_uri: Optional[str] = None
                count: Optional[int] = None

                class Config(BaseConfig):
                    fields = {
                        'items': {
                            'description': f'List of {cls.__name__} that match with query filters'
                        },
                        'next_page_uri': {
                            'description': 'URL to fetch the next page of results'
                        },
                        'count': {
                            'description': f'Counter of {cls.__name__} objects that match with query filters.  \n'
                            f'Included in response only if `count` param was `true`'
                        },
                    }

            QueryResponse.__name__ = f'QueryResponse{cls.__name__}'

            examples = [
                # If param "count" is False return the list of items
                {
                    'summary': 'Query objects',
                    'value': {
                        'items': [response_sample],
                        'next_page_uri': f'{path}?param1=value1&param2=value2',
                    },
                },
                # If param "count" is True return a counter
                {
                    'summary': 'Count objects',
                    'description': 'Sending `true` value in `count` param',
                    'value': {'count': 1},
                },
            ]

            def validate_params(request: Request):
                try:
                    return cls.query_validator(**request.query_params)
                except ValidationError as e:
                    raise UnprocessableEntity(e.json())

            @self.get(
                path,
                summary=f'{cls.__name__} - Query',
                response_model=QueryResponse,
                description=query_description,
                responses=json_openapi(200, 'Successful Response', examples),
                include_in_schema=include_in_schema,
            )
            @copy_attributes(cls)
            async def query(
                query_params: cls.query_validator = Depends(validate_params),  # type: ignore
            ):
                """GET /resource"""
                if self.platform_id_filter_required() and hasattr(
                    cls.model, 'platform_id'
                ):
                    query_params.platform_id = self.current_platform_id

                if self.user_id_filter_required() and hasattr(
                    cls.model, 'user_id'
                ):
                    query_params.user_id = self.current_user_id
                # Call for custom filter implemented in overwritemethod
                self.custom_filter_required(query_params, cls.model)

                filters = cls.get_query_filter(query_params)
                if query_params.count:
                    result = await _count(filters)
                elif hasattr(cls, 'query'):
                    result = await cls.query(
                        await _all(query_params, filters, path)
                    )
                else:
                    result = await _all(query_params, filters, path)
                return result

            async def _count(filters: Q):
                count = await cls.model.objects.filter(filters).async_count()
                return dict(count=count)

            async def _all(query: QueryParams, filters: Q, resource_path: str):
                if query.limit:
                    limit = min(query.limit, query.page_size)
                    query.limit = max(0, query.limit - limit)
                else:
                    limit = query.page_size
                query_set = (
                    cls.model.objects.order_by("-created_at")
                    .filter(filters)
                    .limit(limit)
                )
                items = await query_set.async_to_list()
                item_dicts = [i.to_dict() for i in items]

                has_more: Optional[bool] = None
                if wants_more := query.limit is None or query.limit > 0:
                    # only perform this query if it's necessary
                    has_more = (
                        await query_set.limit(limit + 1).async_count() > limit
                    )

                next_page_uri: Optional[str] = None
                if wants_more and has_more:
                    query.created_before = item_dicts[-1]['created_at']
                    params = query.dict()
                    if self.user_id_filter_required():
                        params.pop('user_id')
                    if self.platform_id_filter_required():
                        params.pop('platform_id')
                    next_page_uri = f'{resource_path}?{urlencode(params)}'
                return dict(items=item_dicts, next_page_uri=next_page_uri)

            return cls

        return wrapper_resource_class


def json_openapi(code: int, description, samples: List[Dict]) -> dict:
    examples = {f'example_{i}': ex for i, ex in enumerate(samples)}
    return {
        code: {
            'description': description,
            'content': {'application/json': {'examples': examples}},
        },
    }



import asyncio
import json
from functools import wraps
from itertools import count
from typing import AsyncGenerator, Callable, Coroutine

from aiobotocore.httpsession import HTTPClientError
from aiobotocore.session import get_session

from ..exc import RetryTask


async def run_task(
    coro: Coroutine,
    sqs,
    queue_url: str,
    receipt_handle: str,
    message_receive_count: int,
    max_retries: int,
) -> None:
    delete_message = True
    try:
        await coro
    except RetryTask:
        delete_message = message_receive_count >= max_retries + 1
    finally:
        if delete_message:
            await sqs.delete_message(
                QueueUrl=queue_url,
                ReceiptHandle=receipt_handle,
            )


async def message_consumer(
    queue_url: str,
    wait_time_seconds: int,
    visibility_timeout: int,
    can_read: asyncio.Event,
    sqs,
) -> AsyncGenerator:
    for _ in count():
        await can_read.wait()
        try:
            response = await sqs.receive_message(
                QueueUrl=queue_url,
                WaitTimeSeconds=wait_time_seconds,
                VisibilityTimeout=visibility_timeout,
                AttributeNames=['ApproximateReceiveCount'],
            )
            messages = response['Messages']
        except KeyError:
            continue
        except HTTPClientError:
            await asyncio.sleep(1)
            continue
        for message in messages:
            yield message


async def get_running_fast_agave_tasks():
    return [
        t
        for t in asyncio.all_tasks()
        if t.get_name().startswith('fast-agave-task')
    ]


def task(
    queue_url: str,
    region_name: str,
    wait_time_seconds: int = 15,
    visibility_timeout: int = 3600,
    max_retries: int = 1,
    max_concurrent_tasks: int = 5,
):
    def task_builder(task_func: Callable):
        @wraps(task_func)
        async def start_task(*args, **kwargs) -> None:
            can_read = asyncio.Event()
            concurrency_semaphore = asyncio.Semaphore(max_concurrent_tasks)
            can_read.set()

            async def concurrency_controller(coro: Coroutine) -> None:
                async with concurrency_semaphore:
                    if concurrency_semaphore.locked():
                        can_read.clear()

                    try:
                        await coro
                    finally:
                        can_read.set()

            session = get_session()
            async with session.create_client('sqs', region_name) as sqs:
                async for message in message_consumer(
                    queue_url,
                    wait_time_seconds,
                    visibility_timeout,
                    can_read,
                    sqs,
                ):
                    body = json.loads(message['Body'])
                    message_receive_count = int(
                        message['Attributes']['ApproximateReceiveCount']
                    )
                    asyncio.create_task(
                        concurrency_controller(
                            run_task(
                                task_func(body),
                                sqs,
                                queue_url,
                                message['ReceiptHandle'],
                                message_receive_count,
                                max_retries,
                            ),
                        ),
                        name='fast-agave-task',
                    )

                # Espera a que terminen todos los tasks pendientes creados por
                # `asyncio.create_task`. De esta forma los tasks
                # podrán borrar el mensaje del queue usando la misma instancia
                # del cliente de SQS
                running_tasks = await get_running_fast_agave_tasks()
                await asyncio.gather(*running_tasks)

        return start_task

    return task_builder



import asyncio
import json
import uuid
from unittest.mock import AsyncMock, call, patch

import aiobotocore.client
import pytest
from aiobotocore.httpsession import HTTPClientError

from fast_agave.exc import RetryTask
from fast_agave.tasks.sqs_tasks import get_running_fast_agave_tasks, task

CORE_QUEUE_REGION = 'us-east-1'


@pytest.mark.asyncio
async def test_execute_tasks(sqs_client) -> None:
    """
    Happy path: Se obtiene el mensaje y se ejecuta el task exitosamente.
    El mensaje debe ser eliminado automáticamente del queue
    """
    test_message = dict(id='abc123', name='fast-agave')

    await sqs_client.send_message(
        MessageBody=json.dumps(test_message),
        MessageGroupId='1234',
    )

    async_mock_function = AsyncMock(return_value=None)
    await task(
        queue_url=sqs_client.queue_url,
        region_name=CORE_QUEUE_REGION,
        wait_time_seconds=1,
        visibility_timeout=1,
    )(async_mock_function)()
    async_mock_function.assert_called_with(test_message)
    assert async_mock_function.call_count == 1

    resp = await sqs_client.receive_message()
    assert 'Messages' not in resp


@pytest.mark.asyncio
async def test_not_execute_tasks(sqs_client) -> None:
    """
    Este caso es cuando el queue está vacío. No hay nada que ejecutar
    """
    async_mock_function = AsyncMock(return_value=None)
    # No escribimos un mensaje en el queue
    await task(
        queue_url=sqs_client.queue_url,
        region_name=CORE_QUEUE_REGION,
        wait_time_seconds=1,
        visibility_timeout=1,
    )(async_mock_function)()
    async_mock_function.assert_not_called()
    resp = await sqs_client.receive_message()
    assert 'Messages' not in resp


@pytest.mark.asyncio
async def test_http_client_error_tasks(sqs_client) -> None:
    """
    Este test prueba el caso cuando hay un error de conexión al intentar
    obtener recibir el mensaje del queue. Se maneja correctamente la
    excepción `HTTPClientError` para evitar que el loop que consume mensajes
    se rompe inesperadamente.
    """

    test_message = dict(id='abc123', name='fast-agave')

    await sqs_client.send_message(
        MessageBody=json.dumps(test_message),
        MessageGroupId='1234',
    )

    original_create_client = aiobotocore.client.AioClientCreator.create_client

    # Esta función hace un patch de la función `receive_message` para simular
    # un error de conexión, la recuperación de la conexión y posteriores
    # recepciones de mensajes sin body del queue.
    async def mock_create_client(*args, **kwargs):
        client = await original_create_client(*args, **kwargs)
        client.receive_message = AsyncMock(
            side_effect=[
                HTTPClientError(error='[Errno 104] Connection reset by peer'),
                await sqs_client.receive_message(),
                dict(),
                dict(),
                dict(),
            ]
        )
        return client

    async_mock_function = AsyncMock(return_value=None)
    with patch(
        'aiobotocore.client.AioClientCreator.create_client', mock_create_client
    ):
        await task(
            queue_url=sqs_client.queue_url,
            region_name=CORE_QUEUE_REGION,
            wait_time_seconds=1,
            visibility_timeout=3,
            max_retries=1,
        )(async_mock_function)()
        async_mock_function.assert_called_once()


@pytest.mark.asyncio
async def test_retry_tasks_default_max_retries(sqs_client) -> None:
    """
    Este test prueba la lógica de reintentos con la configuración default,
    es decir `max_retries=1`

    En este caso el task debe ejecutarse 2 veces
    (la ejecución normal + max_retries)

    Se ejecuta este número de veces para ser consistentes con la lógica
    de reintentos de Celery
    """
    test_message = dict(id='abc123', name='fast-agave')

    await sqs_client.send_message(
        MessageBody=json.dumps(test_message),
        MessageGroupId='1234',
    )

    async_mock_function = AsyncMock(side_effect=RetryTask)

    await task(
        queue_url=sqs_client.queue_url,
        region_name=CORE_QUEUE_REGION,
        wait_time_seconds=1,
        visibility_timeout=1,
    )(async_mock_function)()

    expected_calls = [call(test_message)] * 2
    async_mock_function.assert_has_calls(expected_calls)
    assert async_mock_function.call_count == len(expected_calls)

    resp = await sqs_client.receive_message()
    assert 'Messages' not in resp


@pytest.mark.asyncio
async def test_retry_tasks_custom_max_retries(sqs_client) -> None:
    """
    Este test prueba la lógica de reintentos con la configuración default,
    es decir `max_retries=3`

    En este caso el task debe ejecutarse 4 veces
    (la ejecución normal + max_retries)
    """
    test_message = dict(id='abc123', name='fast-agave')
    await sqs_client.send_message(
        MessageBody=json.dumps(test_message),
        MessageGroupId='1234',
    )

    async_mock_function = AsyncMock(side_effect=RetryTask)

    await task(
        queue_url=sqs_client.queue_url,
        region_name=CORE_QUEUE_REGION,
        wait_time_seconds=1,
        visibility_timeout=1,
        max_retries=3,
    )(async_mock_function)()

    expected_calls = [call(test_message)] * 4
    async_mock_function.assert_has_calls(expected_calls)
    assert async_mock_function.call_count == len(expected_calls)

    resp = await sqs_client.receive_message()
    assert 'Messages' not in resp


@pytest.mark.asyncio
async def test_does_not_retry_on_unhandled_exceptions(sqs_client) -> None:
    """
    Este caso prueba que las excepciones no controladas no se reintentan por
    default (comportamiento consistente con Celery)

    Dentro de task deben manejarse las excepciones esperadas (como desconexión
    de la red). Véase los ejemplos de cómo aplicar este tipo de reintentos
    """
    test_message = dict(id='abc123', name='fast-agave')
    await sqs_client.send_message(
        MessageBody=json.dumps(test_message),
        MessageGroupId='1234',
    )

    async_mock_function = AsyncMock(
        side_effect=Exception('something went wrong :(')
    )

    await task(
        queue_url=sqs_client.queue_url,
        region_name=CORE_QUEUE_REGION,
        wait_time_seconds=1,
        visibility_timeout=1,
        max_retries=3,
    )(async_mock_function)()

    async_mock_function.assert_called_with(test_message)
    assert async_mock_function.call_count == 1

    resp = await sqs_client.receive_message()
    assert 'Messages' not in resp


@pytest.mark.asyncio
async def test_concurrency_controller(
    sqs_client,
) -> None:
    message_id = str(uuid.uuid4())
    test_message = dict(id=message_id, name='fast-agave')
    for i in range(5):
        await sqs_client.send_message(
            MessageBody=json.dumps(test_message),
            MessageGroupId=message_id,
        )

    max_running_tasks = 0

    async def task_counter(_) -> None:
        nonlocal max_running_tasks
        await asyncio.sleep(1)
        running_tasks = len(await get_running_fast_agave_tasks())
        if running_tasks > max_running_tasks:
            max_running_tasks = running_tasks

    await task(
        queue_url=sqs_client.queue_url,
        region_name=CORE_QUEUE_REGION,
        wait_time_seconds=1,
        visibility_timeout=1,
        max_retries=3,
        max_concurrent_tasks=2,
    )(task_counter)()

    assert max_running_tasks == 2
